; FUNCTION FOR COPYING MEMORY BLOCKS
; ESI = SOURCE 32-BIT OFFSET
; EDI = TARGET 32-BIT OFFSET
; ECX = 16-BIT COUNT (64 KB = MAX)
; NOTE #1: INT 15H / FUNCTION 87H IS NOT USED ANY MORE! (FOR COMPATITBILITY)
; NOTE #2: A20 IS ENABLED AT THE BEGINNING OF THE ROUTINE AND DISABLED AT THE END

COPY_BLOCK PROC NEAR
        PUSHFD
        @PUSH           <DS, ES, FS, GS>
        PUSHAD
        MOV             AX, DATA16
        MOV             DS, AX
        MOV             ES, AX

        MOV             AX, WORD PTR [GDTR]
        PUSH            AX
        ADD             WORD PTR [GDTR], 16

        @PUSH           <ESI, EDI, ECX>

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; BEGINNING OF THE GDT SETUP ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     EAX, DS
        SHL     EAX, 4
        ADD     EAX, OFFSET GDT
        MOV     DWORD PTR [GDTR+2], EAX

        MOV     AX, DS
        MOV     DL, AH
        XOR     DH, DH
        SHL     AX, 4
        SHR     DX, 4
        MOV     SI, AX
        MOV     DI, DX

; 16-BIT REAL/PROTECTED MODE MAIN TASK CODE GOES HERE

        LEA     BX, GDT_CS16
        MOV     AX, CS
        XOR     DH, DH
        MOV     DL, AH
        SHL     AX, 4
        SHR     DX, 4
        @SET_GDT_ENTRY

; COMMON DATA SEGMENT GOES HERE

        LEA     BX, GDT_DS16
        MOV     AX, SI
        MOV     DX, DI
        @SET_GDT_ENTRY

; STACK FOR 16/32 REAL/PROTECTED MODE MAIN TASK GOES HERE

        LEA     BX, GDT_SS16
        MOV     AX, SS
        XOR     DH, DH
        MOV     DL, AH
        SHL     AX, 4
        SHR     DX, 4
        @SET_GDT_ENTRY

        @POP            <ECX, EDI, ESI>

;;;;;;;;;;;;;;;;;;;;;;;;;
;; DATA BLOCKS GO HERE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV             GDT_SRC32.BASE_L, SI
        MOV             EAX, ESI
        SHR             EAX, 16
        MOV             GDT_SRC32.BASE_M, AL
        MOV             GDT_SRC32.BASE_H, AH

        MOV             GDT_DST32.BASE_L, DI
        MOV             EAX, EDI
        SHR             EAX, 16
        MOV             GDT_DST32.BASE_M, AL
        MOV             GDT_DST32.BASE_H, AH

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PMODE INITIALIZATION STARTS HERE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        @ENABLE_A20                             ; ENABLE A20 LINE
        IN      AL, PORT_INT_MASK_M
        MOV     [INT_MASK_M], AL                ; SAVE MASK OF HARDWARE INTS AT MASTER CNTRLR
        IN      AL, PORT_INT_MASK_S
        MOV     [INT_MASK_S], AL                ; SAVE MASK OF HARDWARE INTS AT SLAVE CNTRLR
        CLI                                     ; DISABLE MASKABLE INTS
        @DISABLE_NMI                            ; DISABLE NON-MASKABLE INTS

        LGDT    [GDTR]                          ; LOAD GDTR

        MOV     EAX, CR0
        AND     EAX, 01111111111111111111111111110000B
        OR      EAX, 00000000000000000000000000000001B
        MOV     CR0, EAX                        ; SET UP PROTECTED MODE

        @JUMP                                   ; OVERLOAD CODE SELECTOR
        MOV     AX, DS16_SEL
        MOV     DS, AX
        MOV     ES, AX
        MOV     AX, SS16_SEL
        MOV     SS, AX                          ; LOAD SS WITH STACK SELECTOR
        XOR     AX, AX
        MOV     FS, AX
        MOV     GS, AX

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ACTUAL TRANSFERR GOES HERE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     AX, SRC32_SEL
        MOV     DS, AX
        MOV     AX, DST32_SEL
        MOV     ES, AX
        XOR     SI, SI
        XOR     DI, DI
        CLD
        SHR     ECX, 1
        JNC     CB_CONT
        MOVSB
CB_CONT:
        REP     MOVSW

;;;;;;;;;;;;;;;;;;;;;;;;;;
;; BACK TO REAL MODE... ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     AX, DS16_SEL
        MOV     DS, AX
        MOV     ES, AX

        MOV     [ESP32], ESP
        MOV     AX, BIOS_SEL
        MOV     ES, AX
        MOV     WORD PTR ES:[067H], OFFSET CB_SHUTDOWN_RET
        MOV     WORD PTR ES:[069H], CODE16
        MOV     AL, 5
        OUT     PORT_CMOS+1, AL

;        MOV     AL, SHUT_DOWN
;        OUT     PORT_STATUS, AL
        LIDT    [IDTR0]
        INT     0
;CB_RESET:
;        HLT
;        JMP     CB_RESET

CB_SHUTDOWN_RET:
        MOV     AX, DATA16
        MOV     DS, AX
        MOV     ES, AX
        MOV     AX, STACK16
        MOV     SS, AX
        MOV     ESP, [ESP32]
        XOR     AX, AX
        MOV     FS, AX
        MOV     GS, AX

        MOV     AL, [INT_MASK_M]
        OUT     PORT_INT_MASK_M, AL             ; RESTORE MASK OF HARDWARE INTS AT MASTER CNTRLR
        MOV     AL, [INT_MASK_S]
        OUT     PORT_INT_MASK_S, AL             ; RESTORE MASK OF HARDWARE INTS AT SLAVE CNTRLR
        @ENABLE_NMI                             ; ENABLE NON-MASKABLE INTS
        STI                                     ; ENABLE MASKABLE INTS

        @DISABLE_A20                            ; DISABLE A20 LINE

        POP             AX
        MOV             WORD PTR [GDTR], AX
        POPAD
        @POP            <GS, FS, ES, DS>
        POPFD
        RET
COPY_BLOCK ENDP
