;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Detects Real Mode and hardware ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.8086
DETECT  PROC NEAR

;;;;;;;;;;;;;;;;;;;;;;;;
;; DETECTING CPU TYPE ;;
;;;;;;;;;;;;;;;;;;;;;;;;

        CALL    CPU_CHECK
        CMP     AL, 3
        JNC     DETECT_IS_386_PLUS              ; 386+
        LEA     DX, MSG_NEEDED_386
        MOV     AH, 9
        INT     21H
        JMP     DETECT_FAILED

DETECT_IS_386_PLUS:
.386P

;;;;;;;;;;;;;;;;;;;;;;;;;
;; DETECTING REAL MODE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

        SMSW    AX                              ; BUT THIS WORKS PERFECTLY :)
        TEST    AL, 1                           ; IS PROTECTED MODE ACTIVE?
        JZ      DETECT_IS_REAL_MODE             ; NO, IT ISN'T (REAL MODE)
        LEA     DX, MSG_NEEDED_RM
        MOV     AH, 9
        INT     21H
        JMP     DETECT_FAILED                   ; YES, IT IS. EXITING...

DETECT_IS_REAL_MODE:

;;;;;;;;;;;;;;;;;;;;;;;;;
;; DETECTING HIMEM.SYS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     AX, 4300H
        INT     2FH
        CMP     AL, 80H
        JNE     DETECT_NO_HIMEM                 ; HIMEM.SYS IS NOT INSTALLED
        LEA     DX, MSG_FOUND_HIMEM
        MOV     AH, 9
        INT     21H
        JMP     DETECT_FAILED                   ; YES, IT IS. EXITING...

DETECT_NO_HIMEM:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DETECTING VIDEO SYSTEM ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	MOV	AX,1A00H
	INT	10H		; check if VGA type display
	CMP	AL,1AH
	JNE	DETECT_Not_VGA
	CMP	BL,7		; VGA with Monochrome Display?
	JE	DETECT_IS_VGA
	CMP	BL,8            ; VGA with Color Display?
	JE	DETECT_IS_VGA
DETECT_NOT_VGA:
        LEA     DX, MSG_NEEDED_VGA
        MOV     AH, 9
        INT     21H
        JMP     DETECT_FAILED

DETECT_IS_VGA:
        MOV     AX, 1003H
        MOV     BL, 0
        INT     10H             ; DISABLE BLINKING

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COUNTING EXTENDED MEMORY ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        CALL    MEM_COUNT
        MOV     DS:[EXT_MEMORY], EAX    ; bytes above 1MB+64KB

        RET
DETECT_FAILED:
        MOV             AX, 4C00H
        INT             21H
DETECT  ENDP

.8086
CPU_CHECK PROC NEAR
;       8086 CPU check
;       Bits 12-15 are always set on the 8086 processor
;
CC_check_8086:
        pushf                           ;save FLAGS
        pop     bx                      ;store FLAGS in BX
        mov     ax, 0fffh               ;clear bits 12-15
        and     ax, bx                  ;  in FLAGS
        push    ax                      ;store new FLAGS calue on stack
        popf                            ;replace current FLAGS value
        pushf                           ;set new flags
        pop     ax                      ;store new flags in AX
        and     ax, 0f000h              ;if bits 12-15 are set, then CPU
        cmp     ax, 0f000h              ;  is an 8086/8088
        mov     al, 0                   ; save the CPU type to AL
        je      CC_end_cpuid

;
;       Intel 286 CPU check
;       Bits 12-15 are always clear on the Intel processor.
;
check_80286:
.286
        or      bx, 0f000h              ;try to set bits 12-15
        push    bx
        popf
        pushf
        pop     ax
        and     ax, 0f000h              ; if bits 12-15 are cleared,
                                        ;       CPU=Intel 286
        mov     al, 2                   ; turn on Intel 286 Cpu flag
        jz      CC_end_cpuid            ; if CPU is intel 286, check
                                        ; for Intel 287 math coprocessor

        MOV     AL, 3                   ; 386+
.8086
CC_END_CPUID:
        RET
CPU_CHECK ENDP
.386P

MEM_COUNT PROC NEAR
        @ENABLE_A20
        Xor     EAX, EAX
        Mov     AX, CS
        Shl     EAX, 4
        LEA     EDX, MC_GDT
        Add     EAX, EDX
        Mov     DWord Ptr CS:MC_GDTR+2, EAX; Fill in GDTR with physical address
                                        ; of Global Descriptor Table
        @Push   <DS, ES>
        CLI
        LGDT    FWord Ptr CS:MC_GDTR     ; Load Global Descriptor Table
        Mov     EAX, 1
        Mov     CR0, EAX                ; Set Protected Mode
        Mov     AX, 8
        Mov     DS, AX
        Mov     ES, AX
        Mov     FS, AX
        Mov     GS, AX                  ; All segment registers are loaded
                                        ; with 4GB segment's selector
        Xor     EAX, EAX
        Mov     CR0, EAX                ; Set Real Mode
        STI

        Xor     ECX, ECX                ; 0 bytes
        Mov     DS, CX
        Mov     ES, CX
        Mov     ESI, 110000H            ; 1MB + 64KB
        Mov     EDI, ESI
MC_cycle:
        Mov     EAX, DS:[ESI]
        Not     EAX
        Mov     ES:[EDI], EAX
        Mov     EBX, DS:[ESI]
        Cmp     EAX, EBX
        JNE     MC_failed
        Add     ESI, 4096;4
        Mov     EDI, ESI
        Add     ECX, 4096;4
        Jmp     MC_cycle
MC_failed:
;        And     CX, 0FC00H              ; round to 1KB
        Mov     EBX, ECX

;        SHR     ECX, 2
;        MOV     EDI, 110000H
;        XOR     EAX, EAX
;        MOV     ES, AX
;        CLD
;MC_MEMCYC:
;        MOV     ES:[EDI], EAX           ; CLEAR MEMORY
;        ADD     EDI, 4
;        DEC     ECX
;        JNZ     MC_MEMCYC

        @DISABLE_A20
        MOV     EAX, EBX
        @Pop    <ES, DS>
        RET

MC_GDT   DQ      0, 8F92000000FFFFh      ; "0" & "4GB" Descriptors
MC_GDTR  DW      16, 0, 0
MEM_COUNT ENDP
