;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EXAMPLE OF PMODE PAGING STUFF             ;;
;; BY ALEXEI A. FROUNZE 28TH OF MAY, 2000    ;;
;; E-MAIL:   alexfru@chat.ru                 ;;
;; HOMEPAGE: http://alexfru.chat.ru          ;;
;; MIRROR:   http://members.xoom.com/alexfru ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        .MODEL SMALL

        INCLUDE         MACROSES.INC

STACK_SIZE      EQU     1000H

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 16-BIT REAL/PROTECTED MODE CODE SEGMENT ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.8086
CODE16  SEGMENT PARA 'CODE' USE16
CODE16_BEGIN    =       $
ASSUME  CS:CODE16,DS:DATA16,ES:DATA16
MAIN    PROC
        MOV             AX, DATA16
        MOV             DS, AX
        MOV             ES, AX

        LEA             DX, MSG_STARTUP
        MOV             AH, 9
        INT             21H

;;;;;;;;;;;;;;;;;;;;;;;;
;; DETECTING HARDWARE ;;
;;;;;;;;;;;;;;;;;;;;;;;;

        CALL            DETECT
.386P
        LEA             DX, MSG_EXT_MEM
        MOV             AH, 9
        INT             21H
        MOV             EAX, [EXT_MEMORY]
        SHR             EAX, 10
        LEA             DI, WRITE_BUF

        CALL            DEC2STR
        MOV             DWORD PTR [WRITE_BUF+10], 240A0DH

        LEA             DX, WRITE_BUF
        MOV             AH, 9
        INT             21H

        MOV             EAX, 5
        CALL            delayRTC

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WE'RE GOING TO ENTER THE PROTECTED MODE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        CALL            SETUP_GDT

        @ENABLE_A20                             ; ENABLE A20 LINE
        IN      AL, PORT_INT_MASK_M
        MOV     [INT_MASK_M], AL                ; SAVE MASK OF HARDWARE INTS AT MASTER CNTRLR
        IN      AL, PORT_INT_MASK_S
        MOV     [INT_MASK_S], AL                ; SAVE MASK OF HARDWARE INTS AT SLAVE CNTRLR
        CLI                                     ; DISABLE MASKABLE INTS
        @DISABLE_NMI                            ; DISABLE NON-MASKABLE INTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COPYING PAGE TABLE TO THE 110000H ADDRESS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     BX, DS
        MOVZX   EBX, BX
        SHL     EBX, 4

        LEA     SI, PAGETAB
        MOVZX   ESI, SI
        ADD     ESI, EBX
        MOV     EDI, 110000H                    ; 1MB+64KB = PAGE TABLE
        MOV     CX, 1024*4                      ; 1024 PAGE TABLE ENTRIES (4MB)
        CALL    COPY_BLOCK

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COPYING PAGE DIRECTORY TO THE 120000H ADDRESS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        LEA     SI, PAGEDIR
        MOVZX   ESI, SI
        ADD     ESI, EBX
        MOV     EDI, 120000H                    ; 1MB+128KB = PAGE DIRECTORY
        MOV     CX, 4
        CALL    COPY_BLOCK                      ; PAGE DIR IS ALSO A PAGE TAB

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; LOADING CR3 WITH PAGE DIRECTORY ADDRESS (120000H) ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        MOV     EDI, CR3
        AND     EDI, 0FFFH
        OR      EDI, 120000H
        MOV     CR3, EDI                        ; CR3 = ^PAGE DIRECTORY


        MOV     AX, 13H
        INT     10H

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HERE GOES THE SWITCHING ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        LGDT    [GDTR]                          ; LOAD GDTR

        MOV     EAX, CR0
        AND     EAX, 01111111111111111111111111110000B

; THIS DOES NOT ENABLE PAGING AND 16 COLORED BARS WILL BE DRAWN
; IN THIS ORDER (TOP TO DOWN): BLACK, BLUE, ... YELLOW, WHITE
        OR      EAX, 00000000000000000000000000000001B

; THIS ENABLES PAGING AND 16 COLORED BARS WILL BE DRAWN
; IN THIS ORDER (TOP TO DOWN): WHITE, YELLOW, ... BLUE, BLACK
;        OR      EAX, 10000000000000000000000000000001B

        MOV     CR0, EAX                        ; SET UP PROTECTED MODE

; THIS "JMP $" SHOULD HANG COMPUTER IF SET UP OF PMODE IS SUCCESSFUL
;        JMP $

        @JUMP                                   ; OVERLOAD CODE SELECTOR
        MOV     AX, DS16_SEL
        MOV     DS, AX                          ; LOAD DS WITH DATA SELECTOR
        MOV     ES, AX                          ; LOAD ES WITH DATA SELECTOR
        MOV     AX, SS16_SEL
        MOV     SS, AX                          ; LOAD SS WITH STACK SELECTOR
        MOVZX   ESP, SP
        MOV     [ESP32], ESP

        XOR     AX, AX
        MOV     FS, AX                          ; LOAD FS WITH 0
        MOV     GS, AX                          ; LOAD GS WITH 0
        LLDT    AX                              ; LOAD LDTR WITH 0

        @SET_INT_CTRLR 20H, 28H                 ; REINITIALISE MASTER & SLAVE INT CONTROLLERS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             PMODE IS SET UP                              ;;
;;                      HERE MAY BE ANY OWN PMODE CODE                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        PUSH    ES
        MOV     AX, GRAPH_SEL
        MOV     ES, AX
        XOR     DI, DI
        MOV     DX, 0
        CLD
START_GFX_CYCLE:
        MOV     AL, DL
        MOV     CX, 4096
        REP     STOSB
        INC     DX
        CMP     DX, 16
        JC      START_GFX_CYCLE
        POP     ES

        MOV             EAX, 5
        CALL            delayRTC

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WE'RE GOING TO GET BACK TO THE REAL MODE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

START_BACK_TO_16:
        CLI                                     ; DISABLE MASKABLE INTS
        @DISABLE_NMI                            ; DISABLE NON-MASKABLE INTS
        MOV     AX, DS16_SEL
        MOV     DS, AX                          ; SAFETY IS WHAT WE NEED!!!
        MOV     ES, AX                          ; SAFETY IS WHAT WE NEED!!!
        MOV     AX, SS16_SEL
        MOV     SS, AX
        MOV     ESP, [ESP32]

        INCLUDE SET_RM1.INC

        MOV     AL, [INT_MASK_M]
        OUT     PORT_INT_MASK_M, AL             ; RESTORE MASK OF HARDWARE INTS AT MASTER CNTRLR
        MOV     AL, [INT_MASK_S]
        OUT     PORT_INT_MASK_S, AL             ; RESTORE MASK OF HARDWARE INTS AT SLAVE CNTRLR
        @ENABLE_NMI                             ; ENABLE NON-MASKABLE INTS
        STI                                     ; ENABLE MASKABLE INTS
        @DISABLE_A20                            ; DISABLE A20 LINE

        MOV     AX, 3
        INT     10H

        LEA             DX, MSG_CONTACT
        MOV             AH, 9
        INT             21H

        MOV             AX, 4C00H
        INT             21H
MAIN    ENDP

delayRTC PROC NEAR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Input : eax = number of seconds for delay ;;
;; Output: None, registers are preserved     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        PUSH            EAX
        PUSH            EBX
        PUSH            ECX

        MOV             ECX, EAX
        XOR             AL, AL
        OUT             70H, AL
        JMP             DRL1
DRL1:
        IN              AL, 71H
        MOV             BL, AL
DRLL:
        XOR             AL, AL
        OUT             70H, AL
        JMP             DRL2
DRL2:
        IN              AL, 71H
        CMP             AL, BL
        JE              DRLL
        MOV             BL, AL
        LOOP            DRLL

        POP             ECX
        POP             EBX
        POP             EAX
        RET
delayRTC ENDP

        INCLUDE         DETECT.INC
        INCLUDE         NUM2STR.INC
        INCLUDE         SETUPGDT.INC
        INCLUDE         CPYBLCK.INC
CODE16_SIZE     =       $ - CODE16_BEGIN
CODE16  ENDS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COMMON REAL/PROTECTED MODE DATA SEGMENT ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

DATA16  SEGMENT PARA 'DATA'
DATA16_BEGIN    =       $

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; LAUNCH-TIME ERROR MESSAGES ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

MSG_STARTUP     DB      "Example of paging stuff by Alexei A. Frounze (C) 2000",13,10,"$"
MSG_CONTACT     DB      "Contact information:",13,10
                DB      "E-mail  : alexfru@chat.ru",13,10
                DB      "Homepage: http://alexfru.chat.ru",13,10
                DB      "Mirror  : http://members.xoom.com/alexfru",13,10,"$"
MSG_NEEDED_386  DB      "80386 OR BETTER PROCESSOR IS NEEDED.",13,10,"$"
MSG_NEEDED_RM   DB      "PROCESSOR IS ALREADY IN PROTECTED MODE.",13,10
                DB      "PROBABLY WINDOWS OR EMM386.EXE IS LOADED.",13,10
                DB      "LOAD PURE DOS IN REAL MODE.",13,10,"$"
MSG_FOUND_HIMEM DB      "HIMEM.SYS IS INSTALLED. DON'T LOAD IT PLEASE.",13,10,"$"
MSG_NEEDED_VGA  DB      "VGA OR SVGA(VESA COMPATIBLE) VIDEO ADAPTER IS NEEDED.",13,10,"$"
MSG_EXT_MEM     DB      "MEMORY ABOVE 1MB+64KB FOUND (KB): $"
MSG_MEM_ERR     DB      "NOT ENOUGH MEMORY!",13,10,"$"

;;;;;;;;;;;;;;;;;;;;;;
;; STRING CONSTANTS ;;
;;;;;;;;;;;;;;;;;;;;;;

; HEXADECIMAL TABLE FOR CONVERTING BCD TO TEXT
HEX_TAB         DB      "0123456789ABCDEF"

; BUFFER FOR CONVERTING DECIMAL AND HEX NUMBERS TO STRINGS
WRITE_BUF       DB      16 DUP (0), 13, 10, "$"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; GLOBAL DESCRIPTOR TABLE ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        GDTR            DF GDT_SIZE-1

GDT_BEGIN       =       $
GDT     LABEL           WORD
        GDT_0           S_DESC <0,0,0,0,0,0>                            ; 00
        GDT_CS16        S_DESC <CODE16_SIZE-1,,,ACS_CODE,0,>            ; 08
        GDT_DS16        S_DESC <DATA16_SIZE-1,,,ACS_DATA,0,>            ; 10
        GDT_SS16        S_DESC <STACK_SIZE -1,,,ACS_STACK,0,>           ; 18

        GDT_BIOS        S_DESC <SIZE_BIOS-1,LOW_BIOS,HIGH_BIOS,ACS_DATA+ACS_DPL_3,0,0>  ; 20
        GDT_TEXT        S_DESC <SIZE_TEXT-1,LOW_TEXT,HIGH_TEXT,ACS_DATA+ACS_DPL_3,0,0>  ; 28
        GDT_GRAPH       S_DESC <SIZE_GRAPH-1,LOW_GRAPH,HIGH_GRAPH,ACS_DATA+ACS_DPL_3,0,0>;30
        GDT_FLAT        S_DESC <0FFFFH,0,0,ACS_DATA,08FH,0>    ; 4 GB SEGMENT ;))       ; 38
GDT_SIZE        =       $ - GDT_BEGIN
        GDT_SRC32       S_DESC <0FFFFH,,,ACS_DATA,0,>       ; 40 (TEMPORARY)
        GDT_DST32       S_DESC <0FFFFH,,,ACS_DATA,0,>       ; 48 (TEMPORARY)

CS16_SEL        =       GDT_CS16 - GDT_0
DS16_SEL        =       GDT_DS16 - GDT_0
SS16_SEL        =       GDT_SS16 - GDT_0

BIOS_SEL        =       (GDT_BIOS - GDT_0)  OR RPL_3
TEXT_SEL        =       (GDT_TEXT - GDT_0)  OR RPL_3
GRAPH_SEL       =       (GDT_GRAPH - GDT_0) OR RPL_3
FLAT_SEL        =       GDT_FLAT - GDT_0
SRC32_SEL       =       GDT_SRC32 - GDT_0
DST32_SEL       =       GDT_DST32 - GDT_0

;;;;;;;;;;;;;;;;
;; DUMMY IDTR ;;
;;;;;;;;;;;;;;;;

IDTR0           DF      0               ; USED FOR SWITCHING TO REAL MODE
IDTR1           DF      3FFh

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; OTHER SYSTEM VARIABLES ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; ORIGINAL MASKS FOR IRQS
INT_MASK_M      DB      ?
INT_MASK_S      DB      ?

; AMMOUNT OF THE EXTENDED MEMORY ABOVE 1MB+64KB
EXT_MEMORY      DD      ?

; STACK POINTER STORAGE
ESP32           DD      ?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PAGE DIRECTORY THAT IS COPIED TO THE 120000H ADDRESS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PAGEDIR         DD      110003H

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PAGE TABLE THAT IS COPIED TO THE 110000H ADDRESS ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PAGETAB LABEL DWORD
        N = 0
        REPT    160
        DD      N OR 3
        N = N + 4096
        ENDM

;       N = 160*4096
;       REPT    16
;       DD      N OR 3
;       N = N + 4096
;       ENDM

        N = (160+15)*4096       ; REVERSED ORDER OF PAGES IN THE VGA BUFFER
        REPT    16
        DD      N OR 3
        N = N - 4096
        ENDM

        N = (160+16)*4096
        REPT    1024-(160+16)
        DD      N OR 3
        N = N + 4096
        ENDM

DATA16_SIZE     =       $ - DATA16_BEGIN
DATA16  ENDS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; COMMON REAL/PROTECTED MODE STACK SEGMENT ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

STACK16 SEGMENT PARA STACK 'STACK'
                DB      STACK_SIZE DUP(?)
STACK16 ENDS

END     MAIN
